/**
 * laiteudev.c - Udev helper application for naming devices after the
 *               serial number of the underlying file system
 *
 * This file is part of Laite (http://iki.fi/zouppen/laite/)
 * Copyright 2005 Joel Lehtonen <joel dot lehtonen t iki dot fi>
 *
 * Laite is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * See COPYING
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <syslog.h>
#include <errno.h>

#include "common.h"
#include "pako.h"
#include "filesystem.h"

const char* msg_amount = "Invalid amount of arguments: %d (should be %d)";
const char* msg_event = "Udev event is not \"add\" but \"%s\"";
const char* msg_fs = "File system on %s is not recognized";
const char* msg_open = "Cannot open device of %s (%s): %s";
const char* msg_read = "Cannot read boot sector of %s (%s)";
const char* msg_info = "File system on %s is %s, serial number: %s";
const char* msg_nf = "Device name for %s is not found from config";
const char* msg_ok = "Device %s is now assigned to %s";
const char* msg_src = "Logical error in source file %s:%d";
const char* msg_conf_o = "Cannot open config file \"%s\" for reading: %s";
const char* msg_conf_s = "Syntax error in config file at line %d, column %d";
const char* msg_conf_r = "Read error in config file \"%s\" at line %d";
const char* msg_weird = "Mystical error happened: %s";
const char* msg_notudev = "Not in udev environment: ACTION or DEVPATH is not set\n";
const char* msg_help = "To use this udev helper application add the following to udev rules file:\n"
                       "SUBSYSTEM=\"block\", PROGRAM=\"%s %%N configfile\", SYMLINK=\"%%c\"\n";
		       
int main(int argc, char** argv)
{
	char* toiminto = getenv("ACTION");
	char* syspolku = getenv("DEVPATH");

	if ( toiminto == NULL || syspolku == NULL ) {
		// Haetaan polku itseen
		size_t puskPit = 80;
		char pusk[puskPit+1]; // Yksi loppumerkille
		puskPit = readlink("/proc/self/exe", pusk, puskPit);
		if (puskPit == -1) virheErrno(errno,"Ei saada luettua procia");
		pusk[puskPit] = '\0';		

		fprintf(stderr,msg_notudev);
		fprintf(stderr,msg_help, pusk);
		return 1;
	}

	// Oletetaan, ett ollaan udevin koollekutsumia
	
	int left_args = argc-1;
	char **cur_arg = argv+1;
	
	openlog("laite",LOG_PID,LOG_USER);
	
	if ( left_args && !strcmp("-v",*cur_arg) ) {
		verbose = true;
		left_args--; cur_arg++;
	} // Verbose is now set
	
	if ( left_args != 2 ) {
		syslog(LOG_ERR,msg_amount, left_args, 2);
		return 1;
	}

	if ( strcmp("add",toiminto) ) { // jos toiminto ei oikea (add)
		if ( verbose ) syslog(LOG_WARNING,msg_event,toiminto);
		return 2;
	}	

	char* konffi = cur_arg[1];
	char* laite = cur_arg[0];
	
	char sektori[tietoSize];
	const int paluu = lueAlusta(sektori,laite,tietoSize);
	if ( paluu != luku_OK && verbose ) { // switch can't be used with externs :(
		if (paluu == luku_AVAUS)
			syslog(LOG_WARNING,msg_open,syspolku,laite,
			       strerror(errno));
		else if (paluu == luku_LYHYT || paluu == luku_LUKU)
			syslog(LOG_NOTICE,msg_read,syspolku,laite);
		else // error in source
			syslog(LOG_ERR,msg_src,__FILE__,__LINE__);
	}
	if ( paluu != luku_OK ) return 3; // exit anyway
	
	const osiotieto_t* tyyppi = selvitaTyyppi(sektori);
	if ( tyyppi == NULL ) {
		if (verbose) syslog(LOG_NOTICE,msg_fs,syspolku);
		return 4;
	}

	// Let's open the "device configuration" file
	FILE* confFile = fopen(konffi,"r");
	if (confFile == NULL) {
		syslog(LOG_ERR,msg_conf_o,konffi,strerror(errno));
		return 5;
	}

	int found = false;
	int syntaxOK = true;
	int shouldBeEOF = true;
	char* confLine = NULL;
	int lineLen = 0;
	int lineNum = 0;
	
	while ( getline(&confLine, &lineLen, confFile) != -1 ) { // reading file
		const char delim[] = " \t\n";
		int errColumn;

		lineNum++; // For error reporting
		
		// empty row or a comment
		if (confLine[0] == '\n' || confLine[0] == '#') continue;
		
		char* compName = strtok(confLine,delim);
		char* compFs = strtok(NULL,delim);
		char* compSn = strtok(NULL,delim);
		
		if      (compName == NULL) errColumn = 1;
		else if (compFs   == NULL) errColumn = 2;
		else if (compSn   == NULL) errColumn = 3;
		else if (strtok(NULL,delim) != NULL) errColumn = 4;
		else errColumn = 0;
	
		if (errColumn) { // Syntax error in config
			syslog(LOG_ERR,msg_conf_s, lineNum, errColumn);
			shouldBeEOF = false;
			syntaxOK = false;
			break;
		}

		// Luetaan tiedostoa lpi
		if ( strcasecmp(tyyppi->nimi,compFs) ) continue; // fs is uncool

		size_t lenSn = tyyppi->lenSerial;
		char *snBin = sektori + tyyppi->offSerial;
		char snHex[2*lenSn + 1]; // hex takes n*2 bytes plus '\0'
		binhexConv(snHex,snBin,lenSn); // outputs little endian
		
		if ( strcasecmp(snHex,compSn) ) continue; // if sn differs

		// Filesystem type serial numer ok. This is right one.
		syslog(LOG_INFO,msg_ok,syspolku,compName);
		printf("%s\n",compName); // udev gets this
		
		found = true;
		shouldBeEOF = false; // in finland we call this "viri-viri"
		break;
	}
	
	if (shouldBeEOF && !feof(confFile)) syslog(LOG_ERR,msg_conf_r,konffi,lineNum+1);
	if (fclose(confFile)) syslog(LOG_ERR,msg_weird,strerror(errno));
	free(confLine); // buffer allocated by getLine
	
	if (found) return 0; // Happy habby!
	if (!syntaxOK) return 6;

	// Nothing found but config read successful
	
	if (verbose) syslog(LOG_NOTICE,msg_nf,syspolku);
	return 9; // No matching device is found
}

